#include "glfont.h"
#include "SDL_ttf.h"
#include "mzmalloc.h"
#include <string.h>
#include <assert.h>

#include <GL/gl.h>


GlTextBox* gltext_create(const char* text, 
                         TTF_Font* font, 
                         int x, int y,
                         unsigned int rgba) {

  SDL_Color col;
  col.r = ((rgba >> 24) & 0xff);
  col.g = ((rgba >> 16) & 0xff);
  col.b = ((rgba >>  8) & 0xff);

  GlTextBox* rval = mznew(GlTextBox, 1);

  rval->x = x;
  rval->y = y;
  
  SDL_Surface* surface = TTF_RenderText_Blended(font, text, col);

  rval->surface = SDL_CreateRGBSurface(
                                       SDL_SWSURFACE,
                                       surface->w, surface->h,
                                       32,
#if SDL_BYTEORDER == SDL_LIL_ENDIAN /* OpenGL RGBA masks */
                                       0x000000FF, 
                                       0x0000FF00, 
                                       0x00FF0000, 
                                       0xFF000000
#else
                                       0xFF000000,
                                       0x00FF0000, 
                                       0x0000FF00, 
                                       0x000000FF
#endif
                                       );

  assert(rval->surface);

  Uint32 saved_flags;
  Uint8  saved_alpha;

  saved_flags = surface->flags&(SDL_SRCALPHA|SDL_RLEACCELOK);
  saved_alpha = surface->format->alpha;

  if ( (saved_flags & SDL_SRCALPHA) == SDL_SRCALPHA ) {
    SDL_SetAlpha(surface, 0, 0);
  }
  
  /* Copy the surface into the GL texture image */
  SDL_Rect area;
  area.x = 0;
  area.y = 0;
  area.w = surface->w;
  area.h = surface->h;

  SDL_BlitSurface(surface, &area, rval->surface, &area);

  SDL_FreeSurface(surface);

  return rval;

}



void gltext_draw(GlTextBox* tb) {

  assert(tb->surface->w * 4 == tb->surface->pitch);
  
  glRasterPos2i(tb->x, tb->y-tb->surface->h);
  glPixelZoom(1, -1);
  glDrawPixels(tb->surface->w,
               tb->surface->h,
               GL_RGBA,
               GL_UNSIGNED_BYTE,
               tb->surface->pixels);

  
}

void gltext_free(GlTextBox* tb) {
  assert(tb && tb->surface);
  SDL_FreeSurface(tb->surface);
  mzfree(tb);
}
