#ifndef _LEVEL_H_
#define _LEVEL_H_

#include "svgsprite.h"

typedef struct LEVEL Level;
typedef struct HOUSE House;
typedef struct HOUSEFRAME HouseFrame;
typedef struct FLOOR Floor;
typedef struct ROOM Room;
typedef struct ENEMY Enemy;
typedef struct PLAYER Player;

#define ROOM_TYPE_BATH    0x0100
#define ROOM_TYPE_BED     0x0200
#define ROOM_TYPE_KITCHEN 0x0400
#define ROOM_TYPE_LIVING  0x0800
#define ROOM_TYPE_DINING  0x1000
#define ROOM_TYPE_HALL    0x2000
#define ROOM_TYPE_STUDY   0x4000
#define ROOM_TYPE_LAUNDRY 0x8000

#define ROOM_TYPE_MASK    0xff00

#define ROOM_TYPE_SINGLE  (ROOM_TYPE_KITCHEN | ROOM_TYPE_LAUNDRY | ROOM_TYPE_LIVING | ROOM_TYPE_DINING)

#define PROP_FLAG_TREASURE       0x01
#define PROP_FLAG_LIGHTSWITCH    0x02
#define PROP_FLAG_STAIRWAYDOWN   0x04
#define PROP_FLAG_STAIRWAYUP     0x08
#define PROP_FLAG_CANHIDEUNDER   0x10
#define PROP_FLAG_CANHIDEAGAINST 0x20
#define PROP_FLAG_T_TOP          0x40
#define PROP_FLAG_T_CENTER       0x80

#define PROP_FLAG_MASK           0xff
#define PROP_FLAG_SPECIALMASK    0x0f

#define DOG_STATE_SLEEP   0
#define DOG_STATE_WAKEUP  1
#define DOG_STATE_WALK    2
#define DOG_STATE_SCRATCH 3
#define DOG_STATE_BARK    4

#define RESIDENT_STATE_IDLE    0
#define RESIDENT_STATE_WAKEUP  1
#define RESIDENT_STATE_WALK    2
#define RESIDENT_STATE_DESCEND 3 
#define RESIDENT_STATE_ASCEND  4
#define RESIDENT_STATE_SCRATCH 5
#define RESIDENT_STATE_GASP    6

#define PLAYER_STATE_IDLE    0
#define PLAYER_STATE_WALK    1
#define PLAYER_STATE_DESCEND 2
#define PLAYER_STATE_ASCEND  3
#define PLAYER_STATE_STEAL   4
#define PLAYER_STATE_FLIP    5
#define PLAYER_STATE_CROUCH  6
#define PLAYER_STATE_FLATTEN 7

#define LEVEL_RESULT_CONTINUE 0
#define LEVEL_RESULT_ADVANCE  1
#define LEVEL_RESULT_DIE      2
#define LEVEL_RESULT_QUIT     3

#define PLAYER_FLAG_CAN_STEAL   0x01
#define PLAYER_FLAG_CAN_FLIP    0x02
#define PLAYER_FLAG_CAN_CROUCH  0x04
#define PLAYER_FLAG_CAN_FLATTEN 0x08
#define PLAYER_FLAG_CAN_ASCEND  0x10
#define PLAYER_FLAG_CAN_DESCEND 0x20


struct ENEMY {
  Actor* actor;
  Room* room;
  int state;
  double state_time;
  double dx;
  double dy;
  Prop* clip;
};

struct PLAYER {
  Actor* actor;
  Room* room;
  int flashlight_on;
  double flashlight_angle;
  int flags;
  int dx;
  int dy;
  int last_dx;
  int action;
  int has_loot;
  int state;
  double state_time;
  Prop* object;
};

struct LEVEL {
  int level_score;
  unsigned int level;
  int paused;
  double cur_time;
  double last_time;
  double level_total_time;
  double level_time_left;
  House* house;
  Player player;
  Enemy resident;
  Enemy dog;
};

struct ROOM {
  double minx;
  double miny;
  double maxx;
  double maxy;
  ArtVpath wall[6];
  unsigned int type;
  int is_lit;
  int been_lit;
  unsigned int n_props;
  Prop** props;  
};

struct FLOOR {
  unsigned int n_rooms;
  Room** rooms;
  int room_grid[6];
};

struct HOUSEFRAME {
  unsigned int n_paths;
  ArtVpath** paths;
};

struct HOUSE {
  HouseFrame frame;
  double w;
  double h;
  unsigned int cell_w;
  unsigned int single_flags;
  unsigned int n_floors;
  Floor* floors;
  unsigned int n_rooms;
  unsigned int n_treasures;
};


Level* level_generate(ArtLibrary* lib, unsigned int level);
int level_handle_input(Level*);
int level_step_ai(Level*);
void level_draw(ArtLibrary* lib, Level*);
void level_free(Level*);

House* house_generate(ArtLibrary* lib, unsigned int level);
void house_free(House*);

Room* room_create(ArtLibrary* lib, House* h, int type, int flags[3], double minx, double maxx, double minyy, double maxy);
void room_free(Room*);



#endif
