#ifndef _SVGSPRITE_H_
#define _SVGSPRITE_H_

#include "tinysvg.h"
#include "SDL_ttf.h"

typedef struct PROP Prop;
typedef struct ACTOR Actor;
typedef struct SPRITE Sprite;
typedef struct ARTLIBRARY ArtLibrary;
typedef struct FONTSET FontSet;

//////////////////////////////////////////////////////////////////////

struct ARTLIBRARY {
  unsigned int n_svgs;
  TinySvg** svgs;
  unsigned int n_props;
  Prop** props;
  unsigned int n_actors;
  Actor** actors;
  unsigned int n_fonts;
  FontSet** fonts;
};

struct PROP {
  char filename[64];
  TinySvg* svg;
  double x;
  double y;
  unsigned int flags;
  unsigned int value;
};

struct SPRITE {
  TinySvg* svg;
  unsigned int ref_count; // number of actors
  unsigned int n_frames; // number of frames
  double* origin_coords; // 2*n_groups
  double* rotations;     // n_groups * n_frames
};

struct ACTOR {
  char filename[64];
  unsigned int n_sprites;
  Sprite** sprites;
  double x;
  double y;
  double w;
  double h;
  unsigned int lastsprite;
  unsigned int lastf1;
  unsigned int lastf2;
  double lastblend;
};

struct FONTSET {
  char filename[64];
  TTF_Font* small;
  TTF_Font* medium;
  TTF_Font* large;
};


//////////////////////////////////////////////////////////////////////

ArtLibrary* artlibrary_load();
void artlibrary_free(ArtLibrary* l);

TinySvg* artlibrary_lookup_svg(ArtLibrary* l, const char* filename);
Prop* artlibrary_lookup_prop(ArtLibrary* l, const char* filename);
Actor* artlibrary_lookup_actor(ArtLibrary* l, const char* filename);
FontSet* artlibrary_lookup_font(ArtLibrary* l, const char* filename);

//////////////////////////////////////////////////////////////////////

Prop* prop_copy(Prop*);
void prop_free(Prop*);

void prop_draw(Prop*, int invert);
int prop_hoverlap_prop(Prop*, Prop*);

//////////////////////////////////////////////////////////////////////

void sprite_keyframe(Sprite*, 
                     unsigned int f1, unsigned int f2, 
                     double blend);

void sprite_cleartransform(Sprite*);

//////////////////////////////////////////////////////////////////////

Actor* actor_copy(Actor*);
void actor_free(Actor*);

void actor_draw(Actor*, unsigned int sprite,
                unsigned int f1, unsigned int f2, 
                double blend, 
                int flip, int invert);

int actor_overlap_actor(Actor* a1, Actor* a2);
int actor_overlap_prop(Actor* a, Prop* p);

int actor_hoverlap_actor(Actor* a1, Actor* a2);
int actor_hoverlap_prop(Actor* a, Prop* p);

//////////////////////////////////////////////////////////////////////

#endif
